import logging
import random
import time
from datetime import datetime

from normality import stringify
from sqlalchemy.dialects.postgresql import insert as upsert
from sqlalchemy.exc import (
    DatabaseError,
    DisconnectionError,
    IntegrityError,
    OperationalError,
    ResourceClosedError,
    TimeoutError,
)
from sqlalchemy.sql.expression import insert, update

# We have to cast null fragment values to some text to make the
# UniqueConstraint work
DEFAULT_FRAGMENT = "default"
EXCEPTIONS = (
    DatabaseError,
    DisconnectionError,
    OperationalError,
    ResourceClosedError,
    TimeoutError,
)
try:
    from psycopg import DatabaseError, OperationalError

    EXCEPTIONS = (DatabaseError, OperationalError, *EXCEPTIONS)
except ImportError:
    try:
        from psycopg2 import DatabaseError, OperationalError

        EXCEPTIONS = (DatabaseError, OperationalError, *EXCEPTIONS)
    except ImportError:
        pass

log = logging.getLogger(__name__)


class BulkLoader(object):
    def __init__(self, dataset, size):
        self.dataset = dataset
        self.store = dataset.store
        self.size = size
        self.buffer = {}

    def put(self, entity, fragment=None, origin=None):
        origin = origin or self.dataset.origin
        fragment = stringify(fragment) or DEFAULT_FRAGMENT
        if hasattr(entity, "to_dict"):
            entity = entity.to_dict()
        else:
            entity = dict(entity)
        id_ = entity.pop("id")
        if id_:
            self.buffer[(id_, origin, fragment)] = entity
            if len(self.buffer) >= self.size:
                self.flush()
        else:
            log.warning("Entity has no ID!")

    def _store_values(self, conn, values):
        table = self.dataset.table
        try:
            conn.execute(insert(table).values(values))
        except IntegrityError:
            changing = (
                "entity",
                "timestamp",
            )
            for value in values:
                stmt = update(table)
                changed = {c: value.get(c, {}) for c in changing}
                stmt = stmt.values(changed)
                stmt = stmt.where(table.c.id == value["id"])
                stmt = stmt.where(table.c.origin == value["origin"])
                stmt = stmt.where(table.c.fragment == value["fragment"])
                stmt = stmt.where(table.c.timestamp < value["timestamp"])
                conn.execute(stmt)

    def _upsert_values(self, conn, values):
        """Use postgres' upsert mechanism (ON CONFLICT TO UPDATE)."""
        istmt = upsert(self.dataset.table).values(values)
        stmt = istmt.on_conflict_do_update(
            index_elements=["id", "origin", "fragment"],
            set_=dict(
                entity=istmt.excluded.entity,
                timestamp=istmt.excluded.timestamp,
            ),
        )
        conn.execute(stmt)

    def flush(self):
        if not len(self.buffer):
            return
        values = []
        now = datetime.utcnow()
        for (id_, origin, fragment), entity in sorted(self.buffer.items()):
            values.append(
                {
                    "id": id_,
                    "origin": origin,
                    "fragment": fragment,
                    "timestamp": now,
                    "entity": entity,
                }
            )

        for attempt in range(10):
            conn = self.store.engine.connect()
            tx = conn.begin()
            try:
                if self.store.is_postgres:
                    self._upsert_values(conn, values)
                else:
                    self._store_values(conn, values)
                tx.commit()
                conn.close()
                self.buffer = {}
                return
            except EXCEPTIONS:
                tx.rollback()
                conn.close()
                self.dataset.reset()
                self.store.engine.dispose()
                log.exception("Database error storing entities")
                time.sleep(attempt * random.random())



from followthemoney.dataset.dataset import Dataset
from nomenklatura.store import redis_ as nk

from ftmq.store import Store, View


class RedisQueryView(View, nk.RedisView):
    pass


class RedisStore(Store, nk.RedisStore):
    def query(
        self, scope: Dataset | None = None, external: bool = False
    ) -> RedisQueryView:
        scope = scope or self.dataset
        return RedisQueryView(self, scope, external=external)



from sqlalchemy import MetaData, create_engine
from sqlalchemy import inspect as sqlalchemy_inspect

from ftmq.store.fragments.dataset import Fragments
from ftmq.store.fragments.utils import NULL_ORIGIN


class Store(object):
    """A database containing multiple tables that represent
    FtM-store datasets."""

    PREFIX = "ftm"

    def _adjust_psycopg3_uri(self, database_uri: str) -> str:
        """Adjust PostgreSQL URI to use psycopg3 dialect if psycopg is available."""
        if database_uri.startswith(("postgresql://", "postgres://")):
            try:
                import psycopg  # noqa: F401

                # Use psycopg3 dialect for better performance and compatibility
                if database_uri.startswith("postgresql://"):
                    return database_uri.replace(
                        "postgresql://", "postgresql+psycopg://", 1
                    )
                elif database_uri.startswith("postgres://"):
                    return database_uri.replace(
                        "postgres://", "postgresql+psycopg://", 1
                    )
            except ImportError:
                # Fall back to psycopg2 if psycopg3 is not available
                pass
        return database_uri

    def __init__(
        self,
        database_uri: str,
        **config,
    ):
        self.database_uri = self._adjust_psycopg3_uri(database_uri)

        # Configure connection pooling for psycopg3
        config.setdefault("pool_size", 1)
        if self.database_uri.startswith("postgresql+psycopg://"):
            config.setdefault("max_overflow", 5)
            config.setdefault("pool_timeout", 60)
            config.setdefault("pool_recycle", 3600)
            config.setdefault("pool_pre_ping", True)

        self.engine = create_engine(self.database_uri, future=True, **config)
        self.is_postgres = self.engine.dialect.name == "postgresql"
        self.meta = MetaData()

    def get(self, name, origin=NULL_ORIGIN):
        return Fragments(self, name, origin=origin)

    def all(self, origin=NULL_ORIGIN):
        prefix = f"{self.PREFIX}_"
        inspect = sqlalchemy_inspect(self.engine)
        for table in inspect.get_table_names():
            if table.startswith(prefix):
                name = table[len(prefix) :]
                yield Fragments(self, name, origin=origin)

    def close(self):
        self.engine.dispose()

    def __len__(self):
        return len(list(self.all()))

    def __repr__(self):
        return "<Store(%r)>" % self.engine



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", extra="ignore")

    database_uri: str = Field(
        default="sqlite:///ftm_fragments.db", alias="ftm_fragments_uri"
    )



from functools import cache
from typing import Any, Generator, Type

from anystore.types import SDict, StrGenerator
from banal import ensure_list, is_listish
from followthemoney import E, model
from followthemoney.compare import _normalize_names
from followthemoney.dataset import Dataset
from followthemoney.entity import ValueEntity
from followthemoney.proxy import EntityProxy
from followthemoney.schema import Schema
from followthemoney.types import registry
from followthemoney.util import make_entity_id, sanitize_text
from normality import latinize_text, slugify, squash_spaces
from rigour.names import Name, Symbol, tag_org_name, tag_person_name
from rigour.names.tokenize import normalize_name
from rigour.territories import lookup_territory
from rigour.text.scripts import can_latinize

from ftmq.enums import Comparators
from ftmq.types import Entity

DEFAULT_DATASET = "default"


@cache
def make_dataset(name: str | None = DEFAULT_DATASET) -> Dataset:
    name = name or DEFAULT_DATASET
    return Dataset.make({"name": name, "title": name.title()})


@cache
def ensure_dataset(ds: str | Dataset | None = None) -> Dataset:
    if not ds:
        return make_dataset()
    if isinstance(ds, str):
        return make_dataset(ds)
    return ds


@cache
def get_scope_dataset(*names: str) -> Dataset:
    ds = Dataset({"name": "default", "datasets": names})
    ds.children = {make_dataset(n) for n in names}
    return ds


def parse_comparator(key: str) -> tuple[str, Comparators]:
    key, *comparator = key.split("__", 1)
    if comparator:
        comparator = Comparators[comparator[0]]
    else:
        comparator = Comparators["eq"]
    return key, comparator


def parse_unknown_filters(
    filters: tuple[str, ...],
) -> Generator[tuple[str, str | list[str], str], None, None]:
    _filters = (f for f in filters)
    for prop in _filters:
        prop = prop.lstrip("-")
        if "=" in prop:  # 'country=de'
            prop, value = prop.split("=")
        else:  # ("country", "de"):
            value = next(_filters)

        prop, *op = prop.split("__")
        op = op[0] if op else Comparators.eq
        if op == Comparators["in"]:
            # de,fr or ["de", "fr"]
            if is_listish(value):
                value = ensure_list(value)
            else:
                value = value.split(",")
        yield prop, value, op


def make_entity(
    data: SDict,
    entity_type: Type[E] | None = ValueEntity,
    default_dataset: str | Dataset | None = None,
) -> E:
    """
    Create a `Entity` from a json dict.

    Args:
        data: followthemoney data dict that represents entity data.
        entity_type: The entity class to use (`StatementEntity` or `ValueEntity`)
        default_dataset: A default dataset if no dataset in data

    Returns:
        The Entity instance
    """
    etype = entity_type or ValueEntity
    if data.get("id") is None:
        raise ValueError("Entity has no ID.")
    if etype == EntityProxy:
        return EntityProxy.from_dict(data)
    if etype == ValueEntity:
        if not data.get("datasets"):
            dataset = make_dataset(default_dataset).name
            data["datasets"] = [dataset]
        return etype.from_dict(data)
    datasets = data.get("datasets", [])
    if len(datasets) == 1:
        dataset = ensure_dataset(datasets[0])
    else:
        dataset = ensure_dataset(default_dataset)
    return etype.from_data(dataset, data)


def ensure_entity(
    data: dict[str, Any] | Entity | EntityProxy,
    entity_type: Type[E],
    default_dataset: str | Dataset | None = None,
) -> E:
    """
    Ensure input data to be specified `Entity` type

    Args:
        data: entity or data
        entity_type: The entity class to use (`StatementEntity` or `ValueEntity`)
        default_dataset: A default dataset if no dataset in data

    Returns:
        The Entity instance
    """
    if isinstance(data, entity_type):
        if hasattr(data, "datasets"):
            return data
    if isinstance(data, EntityProxy):
        data = data.to_dict()
    return make_entity(data, entity_type, default_dataset)


def apply_dataset(entity: E, dataset: str | Dataset, replace: bool | None = False) -> E:
    dataset = ensure_dataset(dataset)
    data = entity.to_dict()
    if replace:
        data["datasets"] = [dataset.name]
    else:
        data["datasets"].append(dataset.name)
    return make_entity(data, entity.__class__, dataset)


@cache
def get_country_name(code: str) -> str:
    """
    Get the (english) country name for the given 2-letter iso code via
    [rigour.territories](https://rigour.followthemoney.tech/territories/)

    Examples:
        >>> get_country_name("de")
        "Germany"
        >>> get_country_name("xx")
        "xx"
        >>> get_country_name("gb") == get_country_name("uk")
        True  # United Kingdom

    Args:
        code: Two-letter iso code, case insensitive

    Returns:
        Either the country name for a valid code or the code as fallback.
    """
    territory = lookup_territory(code)
    if territory is not None:
        return territory.name
    return code


@cache
def get_country_code(value: Any, splitter: str | None = ",") -> str | None:
    """
    Get the 2-letter iso country code for an arbitrary country name via
    [rigour.territories](https://rigour.followthemoney.tech/territories/)

    Examples:
        >>> get_country_code("Germany")
        "de"
        >>> get_country_code("Deutschland")
        "de"
        >>> get_country_code("Berlin, Deutschland")
        "de"
        >>> get_country_code("Foo")
        None

    Args:
        value: Any input that will be [cleaned][ftmq.util.clean_string]
        splitter: Character to use to get text tokens to find country name for

    Returns:
        The iso code or `None`
    """
    value = clean_string(value)
    if not value:
        return None
    territory = lookup_territory(value)
    if territory is not None:
        return territory.ftm_country
    if splitter:
        for token in value.split(splitter):
            territory = lookup_territory(token.strip())
            if territory is not None:
                return territory.ftm_country
    return None


def join_slug(
    *parts: str | None,
    prefix: str | None = None,
    sep: str = "-",
    strict: bool = True,
    max_len: int = 255,
) -> str | None:
    """
    Create a stable slug from parts with optional validation

    Examples:
        >>> join_slug("foo", "bar")
        "foo-bar"
        >>> join_slug("foo", None, "bar")
        None
        >>> join_slug("foo", None, "bar", strict=False)
        "foo-bar"
        >>> join_slug("foo", "bar", sep="_")
        "foo_bar"
        >>> join_slug("a very long thing", max_len=15)
        "a-very-5c156cf9"

    Args:
        *parts: Multiple (ordered) parts to compute the slug from
        prefix: Add a prefix to the slug
        sep: Parts separator
        strict: Ensure all parts are not `None`
        max_len: Maximum length of the slug. If it exceeds, the returned value
            will get a computed hash suffix

    Returns:
        The computed slug or `None` if validation fails
    """
    sections = [slugify(p, sep=sep) for p in parts]
    if strict and None in sections:
        return None
    texts = [p for p in sections if p is not None]
    if not len(texts):
        return None
    prefix = slugify(prefix, sep=sep)
    if prefix is not None:
        texts = [prefix, *texts]
    slug = sep.join(texts)
    if len(slug) <= max_len:
        return slug
    # shorten slug but ensure uniqueness
    ident = make_entity_id(slug)[:8]
    slug = slug[: max_len - 9].strip(sep)
    return f"{slug}-{ident}"


def get_year_from_iso(value: Any) -> int | None:
    """
    Extract the year from a iso date string or `datetime` object.

    Examples:
        >>>  get_year_from_iso(None)
        None
        >>>  get_year_from_iso("2023")
        2023
        >>>  get_year_from_iso(2020)
        2020
        >>>  get_year_from_iso(datetime.now())
        2024
        >>>  get_year_from_iso("2000-01")
        2000

    Args:
        value: Any input that will be [cleaned][ftmq.util.clean_string]

    Returns:
        The year or `None`
    """
    value = clean_string(value)
    if not value:
        return
    try:
        return int(str(value)[:4])
    except ValueError:
        return


def clean_string(value: Any) -> str | None:
    """
    Convert a value to `None` or a sanitized string without linebreaks

    Examples:
        >>> clean_string(" foo\n bar")
        "foo bar"
        >>> clean_string("foo Bar, baz")
        "foo Bar, baz"
        >>> clean_string(None)
        None
        >>> clean_string("")
        None
        >>> clean_string("  ")
        None
        >>> clean_string(100)
        "100"

    Args:
        value: Any input that will be converted to string

    Returns:
        The cleaned value or `None`
    """
    value = sanitize_text(value)
    if value is None:
        return
    return squash_spaces(value)


def clean_name(value: Any) -> str | None:
    """
    Clean a value and only return it if it is a "name" in the sense of, doesn't
    contain exclusively of special chars

    Examples:
        >>> clean_name("  foo\n Bar")
        "foo Bar"
        >>> clean_name("- - . *")
        None

    Args:
        value: Any input that will be [cleaned][ftmq.util.clean_string]

    Returns:
        The cleaned name or `None`
    """
    value = clean_string(value)
    if slugify(value) is None:
        return
    return value


def make_fingerprint(value: Any) -> str | None:
    """
    Create a stable but simplified string or `None` from input that can be used
    to generate ids (to mimic `fingerprints.generate` which is unstable for IDs
    as its algorithm could change)

    Examples:
        >>> make_fingerprint("Mrs. Jane Doe")
        "doe jane mrs"
        >>> make_fingerprint("Mrs. Jane Mrs. Doe")
        "doe jane mrs"
        >>> make_fingerprint("#")
        None
        >>> make_fingerprint(" ")
        None
        >>> make_fingerprint("")
        None
        >>> make_fingerprint(None)
        None

    Args:
        value: Any input that will be [cleaned][ftmq.util.clean_name]

    Returns:
        The simplified string (fingerprint) or `None` if value is not feasible
            to fingerprint.
    """
    value = clean_name(value)
    if value is None:
        return
    return " ".join(sorted(set(slugify(value).split("-"))))


def entity_fingerprints(entity: EntityProxy) -> set[str]:
    """Get the set of entity name fingerprints, latinized if the alphabet allows
    it and with org / person tags removed depending on entity schema"""
    return make_fingerprints(*entity.names, schemata={entity.schema})


def make_fingerprints(*names: str, schemata: set[Schema] | None = None) -> set[str]:
    """Get the set of name fingerprints, latinized if the alphabet allows
    it and with org / person tags removed depending on given schemata"""
    # FIXME private import
    schemata = schemata or {model["LegalEntity"]}
    fps: set[str] = set()
    for schema in schemata:
        fps.update(set(_normalize_names(schema, names)))
    # add latinized if appropriate
    return {latinize_text(fp) if can_latinize(fp) else fp for fp in fps}


def make_string_id(*values: Any) -> str | None:
    """
    Compute a hash id based on values

    Args:
        *values: Parts to compute id from that will be
            [cleaned][ftmq.util.clean_name]

    Returns:
        The computed hash id or `None` if a parts cleaned value is `None`
    """
    return make_entity_id(*map(clean_name, values))


def make_fingerprint_id(*values: Any) -> str | None:
    """
    Compute a hash id based on values fingerprints

    Args:
        *values: Parts to compute id from that will be
            [fingerprinted][ftmq.util.make_fingerprint]

    Returns:
        The computed hash id or `None` if a parts fingerprinted value is `None`
    """
    return make_entity_id(*map(make_fingerprint, values))


@cache
def prop_is_numeric(schema: Schema, prop: str) -> bool:
    """
    Indicate if the given property is numeric type

    Args:
        schema: followthemoney schema
        prop: Property

    Returns:
        `False` if the property is not numeric type or not found in the schema
            at all
    """
    prop_ = schema.get(prop)
    if prop_ is not None:
        return prop_.type == registry.number
    return False


def get_entity_caption_property(e: Entity) -> SDict:
    """Get the minimal properties dict required to compute the caption"""
    for prop in e.schema.caption:
        if e.caption:
            return {prop: [e.caption]}
        for value in e.get(prop):
            return {prop: [value]}
    return {}


def get_dehydrated_entity(e: Entity) -> Entity:
    """
    Reduce an Entity to only its property dict that is needed to compute the
    caption.
    """
    data = {
        "id": e.id,
        "schema": e.schema.name,
        "properties": get_entity_caption_property(e),
    }
    return make_entity(data, e.__class__)


def get_featured_entity(e: Entity) -> Entity:
    """
    Reduce an Entity with only its featured properties
    """
    featured = get_dehydrated_entity(e)
    for prop in e.schema.featured:
        featured.add(prop, e.get(prop))
    return featured


def must_str(value: Any) -> str:
    value = clean_string(value)
    if not value:
        raise ValueError(f"Value invalid: `{value}`")
    return value


SELECT_SYMBOLS = "__symbols__"
SELECT_ANNOTATED = "__annotated__"


def get_name_symbols(schema: Schema, *names: str) -> set[Symbol]:
    """Get the rigour names symbols for the given schema and list of names"""
    symbols: set[Symbol] = set()
    if schema.is_a("Person"):
        taggers = [tag_person_name]
    elif schema.is_a("Organization"):
        taggers = [tag_org_name]
    elif schema.is_a("LegalEntity"):
        taggers = [tag_org_name, tag_person_name]
    else:
        return symbols
    for name in names:
        n = Name(name)
        for tagger in taggers:
            for symbol in tagger(n, normalize_name).symbols:
                symbols.add(symbol)
    return symbols


def get_symbols(entity: EntityProxy) -> set[Symbol]:
    """Get the rigour names symbols for the given entity"""
    if not entity.schema.is_a("LegalEntity"):
        return set()
    names = entity.get_type_values(registry.name, matchable=True)
    return get_name_symbols(entity.schema, *names)


def inline_symbols(entity: EntityProxy) -> None:
    """Get the rigour names symbols for the given entity and write them to `indexText`"""
    # clean up old symbols from indexText:
    for text in entity.pop("indexText"):
        if not text.startswith(SELECT_SYMBOLS):
            entity.add("indexText", text)
    symbols = get_symbols(entity)
    entity.add("indexText", f"{SELECT_SYMBOLS} {','.join(map(str, symbols))}")


def select_data(e: EntityProxy, prefix: str) -> StrGenerator:
    """Select arbitrary stored data in `indexText` identified by given prefix"""
    for text in e.get("indexText", quiet=True):
        if text.startswith(prefix):
            yield text.replace(prefix, "").strip()


def select_symbols(e: EntityProxy) -> set[str]:
    """Select stored symbols in `indexText`"""
    symbols: set[str] = set()
    for data in select_data(e, SELECT_SYMBOLS):
        symbols.update(data.split(","))
    return symbols


def select_annotations(e: EntityProxy) -> set[str]:
    """Select stored annotations in `indexText`"""
    return {s for s in select_data(e, SELECT_ANNOTATED)}



import json
import logging
from hashlib import sha1
from itertools import count

from normality import stringify

NULL_ORIGIN = "null"

log = logging.getLogger("ftmq.store.fragments")


class StoreException(Exception):
    pass


def safe_fragment(fragment):
    """Make a hashed fragment."""
    fragment = stringify(fragment)
    if fragment is not None:
        fragment = fragment.encode("utf-8", errors="replace")
        return sha1(fragment).hexdigest()


def write_stream(dataset, file, origin=NULL_ORIGIN):
    bulk = dataset.bulk()
    for idx in count(1):
        line = file.readline()
        if not line:
            break
        entity = json.loads(line)
        bulk.put(entity, fragment=str(idx), origin=origin)
        if idx % 10000 == 0:
            log.info("Write [%s]: %s entities", dataset.name, idx)
    bulk.flush()


def iterate_stream(dataset, file, entity_id=None):
    from followthemoney.cli.util import write_entity

    for entity in dataset.iterate(entity_id=entity_id):
        log.debug("[%s]: %s", entity.id, entity.caption)
        write_entity(file, entity)



from typing import Generator, TypeAlias, TypeVar

from followthemoney import Statement, StatementEntity, ValueEntity

# property multi-value
Value: TypeAlias = list[str]
"""FtM property value is always multi-valued string"""

Entity = TypeVar("Entity", StatementEntity, ValueEntity)
"""Generic type used mostly in ftmq"""

# entity generators
Entities: TypeAlias = Generator[Entity, None, None]
"""A generator for generic entity type"""
StatementEntities: TypeAlias = Generator[StatementEntity, None, None]
"""A generator for StatementEntity instances"""
ValueEntities: TypeAlias = Generator[ValueEntity, None, None]
"""A generator for ValueEntity instances"""

# statements
Statements: TypeAlias = Generator[Statement, None, None]
"""A generator for Statement instances"""



"""
https://openaleph.org/docs/lib/ftm-datalake/rfc/#basic-layout

A file-like "datalake" statement store based on parquet files and
[deltalake](https://delta-io.github.io/delta-rs/)

Backend has to be local filesystem, s3 or anything else compatible with
`deltalake`

Layout:
    ```
    ./data/
        _delta_log/
            [ix].json
        bucket=[bucket]/  # things, intervals, documents
            origin=[origin]/
                [uid].parquet
    ```
"""

from functools import cache
from pathlib import Path
from typing import Any, Generator
from urllib.parse import urlparse

import duckdb
import pyarrow as pa
from anystore.interface.lock import Lock
from anystore.logging import get_logger
from anystore.store import Store as FSStore
from anystore.types import SDict
from anystore.util import clean_dict
from deltalake import (
    BloomFilterProperties,
    ColumnProperties,
    DeltaTable,
    WriterProperties,
    write_deltalake,
)
from deltalake._internal import TableNotFoundError
from deltalake.table import FilterConjunctionType
from followthemoney import EntityProxy, StatementEntity, model
from followthemoney.dataset.dataset import Dataset
from followthemoney.statement import Statement
from nomenklatura import settings as nks
from nomenklatura import store as nk
from nomenklatura.db import get_metadata
from pydantic import AliasChoices, Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from sqlalchemy import Boolean, DateTime, column, select, table
from sqlalchemy.sql import Select

from ftmq.query import Query
from ftmq.store.base import DEFAULT_ORIGIN, Store
from ftmq.store.sql import SQLQueryView, SQLStore
from ftmq.types import StatementEntities
from ftmq.util import apply_dataset, ensure_entity, get_scope_dataset

log = get_logger(__name__)

Z_ORDER = ["canonical_id", "prop"]  # don't add more columns here
TARGET_SIZE = 50 * 10_485_760  # 500 MB
PARTITION_BY = ["dataset", "bucket", "origin"]
BUCKET_MENTION = "mention"
BUCKET_PAGE = "page"
BUCKET_PAGES = "pages"
BUCKET_DOCUMENT = "document"
BUCKET_INTERVAL = "interval"
BUCKET_THING = "thing"
STATISTICS_BLOOM = ColumnProperties(
    bloom_filter_properties=BloomFilterProperties(True),
    statistics_enabled="CHUNK",
    dictionary_enabled=True,
)
STATISTICS = ColumnProperties(statistics_enabled="CHUNK", dictionary_enabled=True)
WRITER = WriterProperties(
    data_page_size_limit=64 * 1024,
    dictionary_page_size_limit=512 * 1024,
    max_row_group_size=500_000,
    compression="SNAPPY",
    column_properties={
        "canonical_id": STATISTICS,
        "entity_id": STATISTICS,
        "schema": STATISTICS,
        "prop": STATISTICS_BLOOM,
        "value": STATISTICS_BLOOM,
        "last_seen": ColumnProperties(statistics_enabled="CHUNK"),
    },
)

SA_TO_ARROW: dict[type, pa.DataType] = {
    Boolean: pa.bool_(),
    DateTime: pa.timestamp("us"),
}

TABLE = table(
    nks.STATEMENT_TABLE,
    column("id"),
    column("entity_id"),
    column("canonical_id"),
    column("dataset"),
    column("bucket"),
    column("origin"),
    column("source"),
    column("schema"),
    column("prop"),
    column("prop_type"),
    column("value"),
    column("original_value"),
    column("lang"),
    column("external", Boolean),
    column("first_seen", DateTime),
    column("last_seen", DateTime),
)

ARROW_SCHEMA = pa.schema(
    [(col.name, SA_TO_ARROW.get(type(col.type), pa.string())) for col in TABLE.columns]
)


class StorageSettings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", extra="ignore")

    key: str | None = Field(default=None, alias="aws_access_key_id")
    secret: str | None = Field(default=None, alias="aws_secret_access_key")
    endpoint: str | None = Field(
        default=None,
        validation_alias=AliasChoices("aws_endpoint_url", "fsspec_s3_endpoint_url"),
    )

    @property
    def allow_http(self) -> bool:
        if self.endpoint:
            return not self.endpoint.startswith("https")
        return False

    @property
    def duckdb_endpoint(self) -> str | None:
        if not self.endpoint:
            return
        scheme = urlparse(self.endpoint).scheme
        return self.endpoint[len(scheme) + len("://") :]


storage_settings = StorageSettings()


@cache
def storage_options() -> SDict:
    return clean_dict(
        {
            "AWS_ACCESS_KEY_ID": storage_settings.key,
            "AWS_SECRET_ACCESS_KEY": storage_settings.secret,
            "AWS_ENDPOINT_URL": storage_settings.endpoint,
            "AWS_ALLOW_HTTP": str(storage_settings.allow_http),
            "aws_conditional_put": "etag",
        }
    )


@cache
def setup_duckdb_storage() -> None:
    if storage_settings.secret:
        duckdb.query(f"""CREATE OR REPLACE SECRET secret (
            TYPE s3,
            PROVIDER config,
            KEY_ID '{storage_settings.key}',
            SECRET '{storage_settings.secret}',
            ENDPOINT '{storage_settings.duckdb_endpoint}',
            URL_STYLE 'path',
            USE_SSL '{not storage_settings.allow_http}'
            );""")


@cache
def get_schema_bucket(schema_name: str) -> str:
    s = model[schema_name]
    if s.is_a("Mention"):
        return BUCKET_MENTION
    if s.is_a("Page"):
        return BUCKET_PAGE
    if s.is_a("Pages"):
        return BUCKET_PAGES
    if s.is_a("Document"):
        return BUCKET_DOCUMENT
    if s.is_a("Interval"):
        return BUCKET_INTERVAL
    return BUCKET_THING


def pack_statement(stmt: Statement, source: str | None = None) -> SDict:
    data = stmt.to_db_row()
    data["bucket"] = get_schema_bucket(data["schema"])
    data["source"] = source
    return data


def compile_query(q: Select) -> str:
    table = nks.STATEMENT_TABLE
    sql = str(q.compile(compile_kwargs={"literal_binds": True}))
    return sql.replace(f"FROM {table}", f"FROM arrow as {table}")


class Row:
    """Fake sqlalchemy row-like class"""

    def __init__(self, data: SDict) -> None:
        for key, value in data.items():
            setattr(self, key, value)

    def __iter__(self) -> Generator[Any, None, None]:
        yield from self.__dict__.values()

    def __getitem__(self, i: int) -> Any:
        return list(self.__iter__())[i]


def query_duckdb(q: Select, table: DeltaTable) -> duckdb.DuckDBPyRelation:
    rel = duckdb.arrow(table.to_pyarrow_dataset())
    query = compile_query(q)
    return rel.query("arrow", query)


def stream_duckdb(q: Select, table: DeltaTable) -> Generator[Any, None, None]:
    res = query_duckdb(q, table)
    while rows := res.fetchmany(100_000):
        for row in rows:
            yield Row(dict(zip(res.columns, row)))


def ensure_schema_buckets(q: Query) -> Select:
    if not q.schemata_names:
        return q.sql.statements
    buckets: set[str] = set()
    for schema in q.schemata_names:
        buckets.add(get_schema_bucket(schema))
    return q.sql.statements.where(TABLE.c.bucket.in_(buckets))


class LakeQueryView(SQLQueryView):
    def query(self, query: Query | None = None) -> StatementEntities:
        if query:
            query.table = self.store.table
            query = self.ensure_scoped_query(query)
            sql = ensure_schema_buckets(query)
            yield from self.store._iterate(sql)
        else:
            yield from super().query(query)


class LakeStore(SQLStore[LakeQueryView]):
    def __init__(self, *args, **kwargs) -> None:
        self._backend = FSStore(uri=kwargs.pop("uri"))
        self._partition_by = kwargs.pop("partition_by", PARTITION_BY)
        self._lock: Lock = kwargs.pop("lock", Lock(self._backend))
        self._enforce_dataset = kwargs.pop("enforce_dataset", False)
        kwargs["uri"] = "sqlite:///:memory:"  # fake it till you make it
        get_metadata.cache_clear()
        super().__init__(*args, **kwargs)
        self.table = TABLE
        self.uri = self._backend.uri
        setup_duckdb_storage()

    @property
    def deltatable(self) -> DeltaTable:
        return DeltaTable(self.uri, storage_options=storage_options())

    @property
    def exists(self) -> bool:
        try:
            self.deltatable.version()
            return True
        except TableNotFoundError:
            return False

    def _execute(self, q: Select, stream: bool = True) -> Generator[Any, None, None]:
        if not self.exists:
            return
        yield from stream_duckdb(q, self.deltatable)

    def get_scope(self) -> Dataset:
        if "dataset" not in self._partition_by:
            return super().get_scope()
        names: set[str] = set()
        for child in self._backend._fs.ls(self._backend.uri):
            name = Path(child).name
            if name.startswith("dataset="):
                names.add(name.split("=")[1])
        return get_scope_dataset(*names)

    def view(
        self, scope: Dataset | None = None, external: bool = False
    ) -> LakeQueryView:
        scope = scope or self.dataset
        return LakeQueryView(self, scope, external)

    def writer(
        self, origin: str | None = DEFAULT_ORIGIN, source: str | None = None
    ) -> "LakeWriter":
        return LakeWriter(self, origin=origin or DEFAULT_ORIGIN, source=source)

    def get_origins(self) -> set[str]:
        q = select(self.table.c.origin).distinct()
        return set([r.origin for r in stream_duckdb(q, self.deltatable)])


class LakeWriter(nk.Writer):
    store: LakeStore
    BATCH_STATEMENTS = 1_000_000

    def __init__(
        self,
        store: Store,
        origin: str | None = DEFAULT_ORIGIN,
        source: str | None = None,
    ):
        super().__init__(store)
        self.batch: dict[str, tuple[Statement, str | None]] = {}
        self.origin = origin or DEFAULT_ORIGIN
        self.source = source

    def add_statement(self, stmt: Statement, source: str | None = None) -> None:
        if stmt.entity_id is None:
            return
        stmt.origin = stmt.origin or self.origin
        canonical_id = self.store.linker.get_canonical(stmt.entity_id)
        stmt.canonical_id = canonical_id
        key = f"{canonical_id}\t{stmt.id}"
        self.batch[key] = (stmt, source or self.source)

    def add_entity(
        self,
        entity: EntityProxy,
        origin: str | None = None,
        source: str | None = None,
    ) -> None:
        e = ensure_entity(entity, StatementEntity, self.store.dataset)
        if self.store._enforce_dataset:
            e = apply_dataset(e, self.store.dataset, replace=True)
        for stmt in e.statements:
            if origin:
                stmt.origin = origin
            self.add_statement(stmt, source=source)
        # we check here instead of in `add_statement` as this will keep entities
        # together in the same parquet files
        if len(self.batch) >= self.BATCH_STATEMENTS:
            self.flush()

    def _pack_batches(self) -> Generator[pa.RecordBatch, None, None]:
        batch: list[SDict] = []
        for key in sorted(self.batch):
            stmt, source = self.batch[key]
            batch.append(pack_statement(stmt, source))
            if len(batch) >= 100_000:
                yield pa.RecordBatch.from_pylist(batch, schema=ARROW_SCHEMA)
                batch = []
        if batch:
            yield pa.RecordBatch.from_pylist(batch, schema=ARROW_SCHEMA)

    def flush(self) -> None:
        if self.batch:
            log.info(
                f"Write {len(self.batch)} statements to deltalake ...",
                uri=self.store.uri,
            )
            with self.store._lock:
                reader = pa.RecordBatchReader.from_batches(
                    ARROW_SCHEMA, self._pack_batches()
                )
                write_deltalake(
                    str(self.store.uri),
                    reader,
                    partition_by=self.store._partition_by,
                    mode="append",
                    schema_mode="merge",
                    writer_properties=WRITER,
                    target_file_size=TARGET_SIZE,
                    storage_options=storage_options(),
                    configuration={"delta.enableChangeDataFeed": "true"},
                )

        self.batch = {}

    def pop(self, entity_id: str) -> list[Statement]:
        q = select(TABLE)
        q = q.where(TABLE.c.canonical_id == entity_id)
        statements: list[Statement] = []
        for row in self.store._execute(q):
            statements.append(Statement.from_db_row(row))

        self.store.deltatable.delete(f"canonical_id = '{entity_id}'")
        return statements

    def optimize(
        self,
        vacuum: bool | None = False,
        vacuum_keep_hours: int | None = 0,
        dataset: str | None = None,
        bucket: str | None = None,
        origin: str | None = None,
    ) -> None:
        """
        Optimize the storage: Z-Ordering and compacting

        Args:
            vacuum: Run vacuum after optimization
            vacuum_keep_hours: Retention hours for vacuum
            dataset: Filter optimization to specific dataset partition
            bucket: Filter optimization to specific bucket partition
            origin: Filter optimization to specific origin partition
        """
        filters: FilterConjunctionType = []
        if dataset is not None:
            filters.append(("dataset", "=", dataset))
        if bucket is not None:
            filters.append(("bucket", "=", bucket))
        if origin is not None:
            filters.append(("origin", "=", origin))

        with self.store._lock:
            self.store.deltatable.optimize.z_order(
                Z_ORDER,
                writer_properties=WRITER,
                target_size=TARGET_SIZE,
                partition_filters=filters or None,
            )
            if vacuum:
                self.store.deltatable.vacuum(
                    retention_hours=vacuum_keep_hours,
                    enforce_retention_duration=False,
                    dry_run=False,
                    full=True,
                )



import logging
from contextlib import contextmanager
from datetime import datetime
from typing import Generator, Iterable, TypeAlias

from banal import ensure_list
from followthemoney import EntityProxy, StatementEntity
from followthemoney.dataset.util import dataset_name_check
from normality import slugify
from sqlalchemy import (
    JSON,
    Column,
    Connection,
    DateTime,
    String,
    Table,
    UniqueConstraint,
    distinct,
    func,
    select,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.exc import OperationalError

from ftmq.store.fragments.loader import BulkLoader
from ftmq.store.fragments.utils import NULL_ORIGIN
from ftmq.types import Statements
from ftmq.util import make_dataset

log = logging.getLogger(__name__)
UNDEFINED = (OperationalError,)
try:
    from psycopg.errors import UndefinedTable

    UNDEFINED = (UndefinedTable, *UNDEFINED)
except ImportError:
    try:
        from psycopg2.errors import UndefinedTable

        UNDEFINED = (UndefinedTable, *UNDEFINED)
    except ImportError:
        pass


EntityFragments: TypeAlias = Generator[EntityProxy, None, None]


@contextmanager
def disable_timeout(conn: Connection, store):
    # for long running iterations (e.g. re-index in OpenAleph), for postgres we
    # don't want to get cancelled if a idle_in_transaction_timeout is configured
    # on the server
    if store.is_postgres:
        raw_conn = conn.connection.driver_connection
        with raw_conn.cursor() as cursor:
            cursor.execute("SET idle_in_transaction_session_timeout = 0")
    try:
        yield conn
    finally:
        if store.is_postgres:
            try:
                raw_conn = conn.connection.driver_connection
                with raw_conn.cursor() as cursor:
                    cursor.execute("SET idle_in_transaction_session_timeout = DEFAULT")
            except Exception:
                pass  # Connection might be closed


class Fragments(object):
    def __init__(self, store, name, origin=NULL_ORIGIN):
        self.store = store
        self.name = dataset_name_check(name)
        self.origin = origin
        self._table = None

    @property
    def table(self):
        if self._table is not None:
            return self._table
        table_name = slugify("%s %s" % (self.store.PREFIX, self.name), sep="_")
        if not table_name:
            raise RuntimeError(f"Invalid table name: `{table_name}`")
        json_type = JSONB if self.store.is_postgres else JSON
        self._table = Table(
            table_name,
            self.store.meta,
            Column("id", String, nullable=False),
            Column("origin", String, nullable=False),
            Column("fragment", String, nullable=False),
            Column("timestamp", DateTime, default=datetime.utcnow),
            Column("entity", json_type),
            UniqueConstraint("id", "origin", "fragment"),
            extend_existing=True,
        )
        self._table.create(bind=self.store.engine, checkfirst=True)
        return self._table

    def reset(self):
        self._table = None

    def drop(self):
        log.debug("Dropping ftm-store: %s", self.table)
        try:
            self.table.drop(self.store.engine)
            self.reset()
        except UNDEFINED:
            self.reset()
            raise

    def delete(self, entity_id=None, fragment=None, origin=None):
        table = self.table
        stmt = table.delete()
        if entity_id is not None:
            stmt = stmt.where(table.c.id == entity_id)
        if fragment is not None:
            stmt = stmt.where(table.c.fragment == fragment)
        if origin is not None:
            stmt = stmt.where(table.c.origin == origin)
        try:
            with self.store.engine.connect() as conn:
                conn.execute(stmt)
                conn.commit()
        except UNDEFINED:
            self.reset()
            raise

    def put(self, entity, fragment=None, origin=None):
        bulk = self.bulk()
        bulk.put(entity, fragment=fragment, origin=origin)
        return bulk.flush()

    def bulk(self, size=1000):
        return BulkLoader(self, size)

    def fragments(
        self, entity_ids=None, fragment=None, schema=None, since=None, until=None
    ):
        stmt = self.table.select()
        entity_ids = ensure_list(entity_ids)
        if len(entity_ids) == 1:
            stmt = stmt.where(self.table.c.id == entity_ids[0])
        if len(entity_ids) > 1:
            stmt = stmt.where(self.table.c.id.in_(entity_ids))
        if fragment is not None:
            stmt = stmt.where(self.table.c.fragment == fragment)
        if schema is not None:
            if self.store.is_postgres:
                stmt = stmt.where(self.table.c.entity["schema"].astext == schema)
            else:
                # SQLite JSON support - use json_extract function
                stmt = stmt.where(
                    func.json_extract(self.table.c.entity, "$.schema") == schema
                )
        if since is not None:
            stmt = stmt.where(self.table.c.timestamp >= since)
        if until is not None:
            stmt = stmt.where(self.table.c.timestamp <= until)
        stmt = stmt.order_by(self.table.c.id)
        # stmt = stmt.order_by(self.table.c.origin)
        # stmt = stmt.order_by(self.table.c.fragment)
        conn = self.store.engine.connect()
        try:
            with disable_timeout(conn, self.store) as conn:
                conn = conn.execution_options(stream_results=True)
                for ent in conn.execute(stmt):
                    data = {"id": ent.id, "datasets": [self.name], **ent.entity}
                    if ent.origin != NULL_ORIGIN:
                        data["origin"] = ent.origin
                    yield data
        except Exception:
            self.reset()
            raise
        finally:
            conn.close()

    def partials(
        self, entity_id=None, skip_errors=False, schema=None, since=None, until=None
    ) -> EntityFragments:
        for fragment in self.fragments(
            entity_ids=entity_id, schema=schema, since=since, until=until
        ):
            try:
                yield EntityProxy.from_dict(fragment, cleaned=True)
            except Exception:
                if skip_errors:
                    log.exception("Invalid data [%s]: %s", self.name, fragment["id"])
                    continue
                raise

    def iterate(
        self, entity_id=None, skip_errors=False, schema=None, since=None, until=None
    ) -> EntityFragments:
        if entity_id is None:
            log.info("Using batched iteration for complete dataset.")
            yield from self.iterate_batched(
                skip_errors=skip_errors, schema=schema, since=since, until=until
            )
            return
        entity = None
        invalid = None
        fragments = 1
        for partial in self.partials(
            entity_id=entity_id,
            skip_errors=skip_errors,
            schema=schema,
            since=since,
            until=until,
        ):
            if partial.id == invalid:
                continue
            if entity is not None:
                if entity.id == partial.id:
                    fragments += 1
                    if fragments % 10000 == 0:
                        log.warning(
                            "[%s:%s] aggregated %d fragments...",
                            entity.schema.name,
                            entity.id,
                            fragments,
                        )
                    try:
                        entity.merge(partial)
                    except Exception:
                        if skip_errors:
                            log.exception(
                                "Invalid merge [%s]: %s", self.name, entity.id
                            )
                            invalid = entity.id
                            entity = None
                            fragments = 1
                            continue
                        raise
                    continue
                yield entity
            entity = partial
            fragments = 1
        if entity is not None:
            yield entity

    def iterate_batched(
        self, skip_errors=False, batch_size=10_000, schema=None, since=None, until=None
    ) -> EntityFragments:
        """
        For large datasets an overall sort is not feasible, so we iterate in
        sorted batched IDs.
        """
        for entity_ids in self.get_sorted_id_batches(
            batch_size, schema=schema, since=since, until=until
        ):
            yield from self.iterate(
                entity_id=entity_ids,
                skip_errors=skip_errors,
                schema=schema,
                since=since,
                until=until,
            )

    def get_sorted_id_batches(
        self, batch_size=10_000, schema=None, since=None, until=None
    ) -> Generator[list[str], None, None]:
        """
        Get sorted ID batches to speed up iteration and useful to parallelize
        processing of iterator Entities
        """
        last_id = None
        while True:
            stmt = select(self.table.c.id).distinct()
            if last_id is not None:
                stmt = stmt.where(self.table.c.id > last_id)
            if schema is not None:
                if self.store.is_postgres:
                    stmt = stmt.where(self.table.c.entity["schema"].astext == schema)
                else:
                    # SQLite JSON support - use json_extract function
                    stmt = stmt.where(
                        func.json_extract(self.table.c.entity, "$.schema") == schema
                    )
            if since is not None:
                stmt = stmt.where(self.table.c.timestamp >= since)
            if until is not None:
                stmt = stmt.where(self.table.c.timestamp <= until)
            stmt = stmt.order_by(self.table.c.id).limit(batch_size)
            try:
                with self.store.engine.connect() as conn:
                    res = conn.execute(stmt)
                    entity_ids = [r.id for r in res.fetchall()]
            except Exception:
                self.reset()
                raise
            if not entity_ids:
                return
            yield entity_ids
            last_id = entity_ids[-1]

    def get_sorted_ids(
        self, batch_size=10_000, schema=None, since=None, until=None
    ) -> Generator[str, None, None]:
        """Get sorted IDs, optionally filtered by schema"""
        for batch in self.get_sorted_id_batches(batch_size, schema, since, until):
            yield from batch

    def statements(
        self,
        entity_ids: Iterable[str] | None = None,
        origin: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
    ) -> Statements:
        """Iterate unsorted statements with its fragment origins"""
        stmt = self.table.select()
        entity_ids = ensure_list(entity_ids)
        if len(entity_ids) == 1:
            stmt = stmt.where(self.table.c.id == entity_ids[0])
        if len(entity_ids) > 1:
            stmt = stmt.where(self.table.c.id.in_(entity_ids))
        if origin is not None:
            stmt = stmt.where(self.table.c.origin == origin)
        if since is not None:
            stmt = stmt.where(self.table.c.timestamp >= since)
        if until is not None:
            stmt = stmt.where(self.table.c.timestamp <= until)
        conn = self.store.engine.connect()
        default_dataset = make_dataset(self.name)
        try:
            conn = conn.execution_options(stream_results=True)
            for fragment in conn.execute(stmt):
                data = {"id": fragment.id, "datasets": [self.name], **fragment.entity}
                entity = StatementEntity.from_dict(
                    data, default_dataset=default_dataset
                )
                for statement in entity.statements:
                    statement.last_seen = fragment.timestamp.isoformat()
                    statement.origin = (
                        fragment.origin if fragment.origin != NULL_ORIGIN else None
                    )
                    yield statement
        except Exception:
            self.reset()
            raise
        finally:
            conn.close()

    def get(self, entity_id) -> EntityProxy | None:
        for entity in self.iterate(entity_id=entity_id):
            return entity

    def __iter__(self) -> EntityFragments:
        return self.iterate()

    def __len__(self):
        q = select(func.count(distinct(self.table.c.id)))
        with self.store.engine.connect() as conn:
            res = conn.execute(q)

        return res.scalar()

    def __repr__(self):
        return "<Dataset(%r, %r)>" % (self.store, self.name)



import os
from collections import defaultdict
from decimal import Decimal
from typing import Generic

from anystore.util import clean_dict
from followthemoney.dataset.dataset import Dataset
from nomenklatura.db import get_metadata
from nomenklatura.store import sql as nk
from sqlalchemy import select
from typing_extensions import TypeVar

from ftmq.aggregations import AggregatorResult
from ftmq.enums import Fields
from ftmq.model.stats import Collector, DatasetStats
from ftmq.query import Query
from ftmq.store.base import Store, View
from ftmq.types import StatementEntities
from ftmq.util import get_scope_dataset

V = TypeVar("V", bound=View, default="SQLQueryView")

MAX_SQL_AGG_GROUPS = int(os.environ.get("MAX_SQL_AGG_GROUPS", 10))


def clean_agg_value(value: str | Decimal) -> str | float | int | None:
    if isinstance(value, Decimal):
        return float(value)
    return value


class SQLQueryView(View, nk.SQLView):
    def ensure_scoped_query(self, query: Query) -> Query:
        if not query.datasets:
            return query.where(dataset__in=self.dataset_names)
        if query.dataset_names - self.dataset_names:
            raise ValueError("Query datasets outside view scope")
        return query

    def query(self, query: Query | None = None) -> StatementEntities:
        if query:
            query = self.ensure_scoped_query(query)
            yield from self.store._iterate(query.sql.statements)
        else:
            view = self.store.view(self.scope)
            yield from view.entities()

    def stats(self, query: Query | None = None) -> DatasetStats:
        query = self.ensure_scoped_query(query or Query())
        key = f"stats-{hash(query)}"
        if key in self._cache:
            return self._cache[key]

        c = Collector()
        for schema, count in self.store._execute(query.sql.things, stream=False):
            c.things[schema] = count
        for schema, count in self.store._execute(query.sql.intervals, stream=False):
            c.intervals[schema] = count
        for country, count in self.store._execute(
            query.sql.things_countries, stream=False
        ):
            if country is not None:
                c.things_countries[country] = count
        for country, count in self.store._execute(
            query.sql.intervals_countries, stream=False
        ):
            if country is not None:
                c.intervals_countries[country] = count

        stats = c.export()
        for start, end in self.store._execute(query.sql.date_range, stream=False):
            if start:
                stats.start = start
            if end:
                stats.end = end
            break

        stats.entity_count = self.count(query)
        self._cache[key] = stats
        return stats

    def count(self, query: Query | None = None) -> int:
        if query is not None:
            for res in self.store._execute(query.sql.count, stream=False):
                for count in res:
                    return count
        return 0

    def aggregations(self, query: Query) -> AggregatorResult | None:
        if not query.aggregations:
            return
        query = self.ensure_scoped_query(query)
        key = f"agg-{hash(query)}"
        if key in self._cache:
            return self._cache[key]
        res: AggregatorResult = defaultdict(dict)

        for prop, func, value in self.store._execute(
            query.sql.aggregations, stream=False
        ):
            res[func][prop] = clean_agg_value(value)

        if query.sql.group_props:
            res["groups"] = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
            for prop in query.sql.group_props:
                if prop == Fields.year:
                    start, end = self.stats(query).years
                    if start or end:
                        groups = range(start or end, (end or start) + 1)
                    else:
                        groups = []
                else:
                    groups = [
                        r[0]
                        for r in self.store._execute(
                            query.sql.get_group_counts(prop, limit=MAX_SQL_AGG_GROUPS),
                            stream=False,
                        )
                    ]
                for group in groups:
                    for agg_prop, func, value in self.store._execute(
                        query.sql.get_group_aggregations(prop, group), stream=False
                    ):
                        res["groups"][prop][func][agg_prop][group] = clean_agg_value(
                            value
                        )
        res = clean_dict(res)
        self._cache[key] = res
        return res


class SQLStore(Store[V], nk.SQLStore, Generic[V]):
    def __init__(self, *args, **kwargs) -> None:
        get_metadata.cache_clear()  # FIXME
        super().__init__(*args, **kwargs)

    def get_scope(self) -> Dataset:
        q = select(self.table.c.dataset).distinct()
        names: set[str] = set()
        for row in self._execute(q, stream=False):
            names.add(row[0])
        return get_scope_dataset(*names)

    def view(self, scope: Dataset | None = None, external: bool = False) -> V:
        scope = scope or self.dataset
        return SQLQueryView(self, scope, external=external)  # type: ignore[return-value]



from followthemoney.dataset.dataset import Dataset
from nomenklatura.store import memory as nk

from ftmq.store.base import Store, View
from ftmq.util import get_scope_dataset


class MemoryQueryView(View, nk.MemoryView):
    pass


class MemoryStore(Store, nk.MemoryStore):
    def get_scope(self) -> Dataset:
        return get_scope_dataset(*self.entities.keys())

    def view(self, scope: Dataset | None = None, external: bool = False) -> View:
        scope = scope or self.dataset
        return MemoryQueryView(self, scope, external=external)



from ftmq.io import smart_read_proxies, smart_write_proxies
from ftmq.query import Query
from ftmq.util import make_entity

__version__ = "4.5.5"
__all__ = [
    "smart_read_proxies",
    "smart_write_proxies",
    "Query",
    "make_entity",
]



from followthemoney.dataset.dataset import Dataset
from nomenklatura.store import level as nk

from ftmq.store.base import Store, View
from ftmq.util import get_scope_dataset


class LevelDBQueryView(View, nk.LevelDBView):
    pass


class LevelDBStore(Store, nk.LevelDBStore):
    def get_scope(self) -> Dataset:
        names: set[str] = set()
        with self.db.iterator(prefix=b"s:", include_value=False) as it:
            for k in it:
                dataset = k.decode().split(":")[3]
                names.add(dataset)
        return get_scope_dataset(*names)

    def view(
        self, scope: Dataset | None = None, external: bool = False
    ) -> LevelDBQueryView:
        scope = scope or self.dataset
        return LevelDBQueryView(self, scope, external=external)



from functools import cached_property
from typing import Any, Self

from alephclient.api import AlephAPI
from alephclient.settings import API_KEY, HOST, MAX_TRIES
from followthemoney import Dataset, EntityProxy, Statement, StatementEntity, ValueEntity
from followthemoney.namespace import Namespace
from furl import furl
from nomenklatura.resolver import Resolver
from nomenklatura.store import Writer

from ftmq.query import Query
from ftmq.store.base import Store, View
from ftmq.types import StatementEntities
from ftmq.util import apply_dataset, ensure_entity, make_entity

uns = Namespace()


def parse_uri(uri: str) -> tuple[str, str | None, str | None]:
    """
    http://host.org
    https://dataset@host.org
    https://dataset:api_key@host.org
    """
    u = furl(uri)
    api_key = str(u.password) if u.password else API_KEY
    dataset = str(u.username) if u.username else None
    return f"{u.scheme}://{u.host}", api_key, dataset


class AlephStore(Store):
    def __init__(
        self, host: str | None = None, api_key: str | None = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.host = host or HOST
        self.api_key = api_key or API_KEY

    @classmethod
    def from_uri(
        cls,
        uri: str,
        dataset: Dataset | str | None = None,
        linker: Resolver | None = None,
    ) -> Self:
        host, api_key, foreign_id = parse_uri(uri)
        if dataset is None and foreign_id is not None:
            dataset = foreign_id

        return cls(dataset=dataset, linker=linker, host=host, api_key=api_key)

    @cached_property
    def api(self):
        return AlephAPI(self.host, self.api_key, retries=MAX_TRIES)

    @cached_property
    def collection(self) -> dict[str, Any]:
        return self.api.load_collection_by_foreign_id(self.dataset.name)

    @cached_property
    def ns(self) -> Namespace:
        return Namespace(self.dataset.name)

    def view(
        self, scope: Dataset | None = None, external: bool = False
    ) -> "AlephQueryView":
        return AlephQueryView(self, scope, external=external)

    def writer(self) -> "AlephWriter":
        return AlephWriter(self)


class AlephQueryView(View):
    store: AlephStore

    def entities(self, query: Query | None = None) -> StatementEntities:
        for proxy in self.store.api.stream_entities(self.store.collection):
            proxy = make_entity(
                proxy, StatementEntity, default_dataset=self.store.dataset.name
            )
            yield uns.apply(proxy)

    def get_entity(self, id: str) -> StatementEntity | None:
        entity_id = self.store.ns.sign(id)
        if not entity_id:
            raise RuntimeError(f"Invalid id: `{id}`")
        proxy = self.store.api.get_entity(entity_id)
        if proxy is not None:
            proxy = make_entity(proxy, StatementEntity, self.store.dataset.name)
            return uns.apply(proxy)
        return None


class AlephWriter(Writer):
    BATCH = 1_000
    store: AlephStore

    def __init__(self, store: AlephStore):
        self.store = store
        self.batch: list[ValueEntity] = []

    def flush(self) -> None:
        if self.batch:
            self.store.api.write_entities(self.store.collection["id"], self.batch)
        self.batch = []

    def add_entity(self, entity: EntityProxy) -> None:
        e = ensure_entity(entity, ValueEntity, self.store.dataset)
        e = apply_dataset(e, dataset=self.store.dataset)
        e.id = self.store.linker.get_canonical(e.id)
        self.batch.append(e)
        if len(self.batch) >= self.BATCH:
            self.flush()

    def pop(self, entity_id: str) -> list[Statement]:
        # FIXME this actually doesn't delete anything
        self.flush()
        statements: list[Statement] = []
        view = self.store.default_view()
        entity = view.get_entity(entity_id)
        if entity is not None:
            statements = list(entity.statements)
        return statements



from typing import Any, Iterable, TypeVar, Union

from banal import as_bool, ensure_list, is_listish
from followthemoney import model
from followthemoney.property import Property
from followthemoney.schema import Schema
from followthemoney.types import registry

from ftmq.enums import Comparators
from ftmq.types import Entity, Value


class Lookup:
    IN = Comparators["in"]
    EQUALS = Comparators["eq"]
    NULL = Comparators["null"]

    def __init__(self, comparator: Comparators, value: Value | None = None):
        self.comparator = self.get_comparator(comparator)
        self.value = value

    def __str__(self) -> str:
        return str(self.comparator)

    def __eq__(self, other: Any) -> bool:
        return str(self) == str(other)

    def get_comparator(self, comparator: str) -> Comparators:
        try:
            return Comparators[comparator]
        except KeyError:
            raise ValueError(f"Invalid oparator: `{comparator}`")

    def apply(self, value: str | None) -> bool:
        if self.comparator == "eq":
            return value == self.value
        if self.comparator == "not":
            return value != self.value
        if self.comparator == "in":
            return value in self.value
        if self.comparator == "not_in":
            return value not in self.value
        if self.comparator == "startswith":
            return value.startswith(self.value)
        if self.comparator == "endswith":
            return value.endswith(self.value)
        if self.comparator == "null":
            return not value == self.value
        if self.comparator == "gt":
            return value > self.value
        if self.comparator == "gte":
            return value >= self.value
        if self.comparator == "lt":
            return value < self.value
        if self.comparator == "lte":
            return value <= self.value
        if self.comparator == "like":
            return self.value in value
        if self.comparator == "ilike":
            return self.value.lower() in value.lower()
        return False


class BaseFilter:
    def __init__(
        self,
        value: Value,
        comparator: Comparators | None = None,
    ):
        try:
            self.comparator = Comparators[comparator or "eq"]
        except KeyError:
            raise ValueError(f"Invalid comparator `{comparator}`")
        self.value: Value = self.get_casted_value(value)
        self.lookup: Lookup = Lookup(self.comparator, self.value)

    def __hash__(self) -> int:
        return hash((self.key, str(self.lookup), str(self.value)))

    def __eq__(self, other: Any) -> bool:
        return hash(self) == hash(other)

    def __lt__(self, other: Any) -> bool:
        # allow ordering (helpful for testing)
        return hash(self) < hash(other)

    def __gt__(self, other: Any) -> bool:
        # allow ordering (helpful for testing)
        return hash(self) > hash(other)

    def to_dict(self) -> dict[str, Any]:
        if self.comparator == Lookup.EQUALS:
            key = self.key
        else:
            key = f"{self.key}__{self.lookup}"
        return {key: self.value}

    def apply(self, entity: Entity) -> bool:
        return self.lookup.apply(self.value)

    def get_casted_value(self, value: Any) -> Value:
        if self.comparator == Lookup.IN:
            return set([self.stringify(v) for v in ensure_list(value)])
        if self.comparator == Lookup.NULL:
            return as_bool(value)
        if is_listish(value):
            raise ValueError(f"Invalid value for `{self.comparator}`: {value}")
        return self.stringify(value) if value is not None else None

    def stringify(self, value: Any) -> str:
        if hasattr(value, "name"):
            return value.name
        return str(value)


class DatasetFilter(BaseFilter):
    key = "dataset"

    def apply(self, entity: Entity) -> bool:
        if self.comparator == Lookup.EQUALS:
            return self.value in entity.datasets
        for value in entity.datasets:
            if self.lookup.apply(value):
                return True
        return False


class OriginFilter(BaseFilter):
    key = "origin"

    def apply(self, entity: Entity) -> bool:
        if not hasattr(entity, "context"):
            return False
        origins = ensure_list(entity.context.get("origin"))
        if self.comparator == Lookup.EQUALS:
            return self.value in origins
        for value in origins:
            if self.lookup.apply(value):
                return True
        return False


class SchemaFilter(BaseFilter):
    key = "schema"

    def __init__(
        self,
        value: Value | Schema | Iterable[Schema],
        comparator: Comparators | None = None,
        schema_include_descendants: bool = False,
        schema_include_matchable: bool = False,
    ):
        super().__init__(value, comparator)
        self.schemata: set[Schema] = set()
        for schema in ensure_list(value):
            schema = model.get(schema)
            if schema is not None:
                self.schemata.add(schema)
                if schema_include_descendants:
                    self.schemata.update(schema.descendants)
                if schema_include_matchable:
                    self.schemata.update(schema.matchable_schemata)
        if not self.schemata and self.comparator in ("eq", "in", "not", "not_in"):
            raise ValueError(f"Invalid schema: `{self.value}`")
        if len(self.schemata) > 1:
            if self.comparator not in ("eq", "in", "not", "not_in"):
                raise ValueError(
                    f"Invalid schema lookup: `{self.comparator}` with multiple schemata"
                )
            self.value = {s.name for s in self.schemata}
            if self.comparator == Comparators["eq"]:
                self.comparator = Comparators["in"]
            elif self.comparator == Comparators["not"]:
                self.comparator = Comparators["not_in"]

    def apply(self, entity: Entity) -> bool:
        if len(self.schemata) > 1:
            return entity.schema in self.schemata
        return self.lookup.apply(entity.schema.name)


class PropertyFilter(BaseFilter):
    def __init__(self, prop: Property, value: Value, comparator: str | None = None):
        super().__init__(value, comparator)
        self.key = self.validate(prop)

    def apply(self, entity: Entity) -> bool:
        for value in entity.get(self.key, quiet=True):
            if self.lookup.apply(value):
                return True
        return False

    def validate(self, prop: str | Property) -> str:
        if isinstance(prop, Property):
            return prop.name
        if isinstance(prop, str):
            for p in model.properties:
                if p.name == prop or p.qname == prop:
                    return prop
        raise ValueError(f"Invalid prop: `{prop}`")


class ReverseFilter(BaseFilter):
    """
    Filter for entities that point to a given entity (via id)
    """

    key = "reverse"

    def apply(self, entity: Entity) -> bool:
        for prop, value in entity.itervalues():
            if prop.type == registry.entity:
                if self.lookup.apply(value):
                    return True
        return False


class IdFilter(BaseFilter):
    key = "id"

    def apply(self, entity: Entity) -> bool:
        return self.lookup.apply(entity.id)


class EntityIdFilter(IdFilter):
    key = "entity_id"


class CanonicalIdFilter(IdFilter):
    key = "canonical_id"


Filter = Union[
    DatasetFilter,
    SchemaFilter,
    PropertyFilter,
    ReverseFilter,
    EntityIdFilter,
    CanonicalIdFilter,
]
F = TypeVar("F", bound=Filter)

FILTERS = {
    "dataset": DatasetFilter,
    "schema": SchemaFilter,
    "property": PropertyFilter,
    "reverse": ReverseFilter,
    "entity_id": EntityIdFilter,
    "canonical_id": CanonicalIdFilter,
    "origin": OriginFilter,
}



from functools import cache

from ftmq.store.fragments.dataset import Fragments
from ftmq.store.fragments.settings import Settings
from ftmq.store.fragments.store import Store
from ftmq.store.fragments.utils import NULL_ORIGIN


@cache
def get_store(database_uri: str | None = None, **config) -> Store:
    settings = Settings()
    return Store(database_uri=database_uri or settings.database_uri, **config)


@cache
def get_fragments(
    name: str,
    origin: str | None = NULL_ORIGIN,
    database_uri: str | None = None,
    **config,
) -> Fragments:
    settings = Settings()
    uri = database_uri or settings.database_uri
    store = get_store(uri, **config)
    return store.get(name, origin=origin or NULL_ORIGIN)



from functools import cache
from pathlib import Path
from urllib.parse import urlparse

from anystore.types import Uri
from followthemoney.dataset.dataset import Dataset
from nomenklatura import Resolver, settings
from nomenklatura.db import get_metadata

from ftmq.store.base import Store, View
from ftmq.store.memory import MemoryStore


@cache
def get_store(
    uri: Uri | None = settings.DB_URL,
    dataset: Dataset | str | None = None,
    linker: Resolver | None = None,
) -> Store:
    """
    Get an initialized [Store][ftmq.store.base.Store]. The backend is inferred
    by the scheme of the store uri.

    Example:
        ```python
        from ftmq.store import get_store

        # an in-memory store:
        get_store("memory://")

        # a leveldb store:
        get_store("leveldb:///var/lib/data")

        # a redis (or kvrocks) store:
        get_store("redis://localhost")

        # a sqlite store
        get_store("sqlite:///data/followthemoney.db")
        ```

    Args:
        uri: The store backend uri
        dataset: A `followthemoney.Dataset` instance to limit the scope to
        linker: A `nomenklatura.Resolver` instance with linked / deduped data

    Returns:
        The initialized store. This is a cached object.
    """
    uri = str(uri)
    parsed = urlparse(uri)
    if parsed.scheme == "memory":
        return MemoryStore(dataset, linker=linker)
    if parsed.scheme == "leveldb":
        path = uri.replace("leveldb://", "")
        path = Path(path).absolute()
        try:
            from ftmq.store.level import LevelDBStore

            return LevelDBStore(dataset, path=path, linker=linker)
        except ImportError:
            raise ImportError("Can not load LevelDBStore. Install `plyvel`")
    if parsed.scheme == "redis":
        try:
            from ftmq.store.redis import RedisStore

            return RedisStore(dataset, linker=linker)
        except ImportError:
            raise ImportError("Can not load RedisStore. Install `redis`")
    if "sql" in parsed.scheme:
        try:
            from ftmq.store.sql import SQLStore

            get_metadata.cache_clear()
            return SQLStore(dataset, uri=uri, linker=linker)
        except ImportError:
            raise ImportError("Can not load SqlStore. Install sql dependencies.")
    if "aleph" in parsed.scheme:
        try:
            from ftmq.store.aleph import AlephStore

            return AlephStore.from_uri(uri, dataset=dataset, linker=linker)
        except ImportError:
            raise ImportError("Can not load AlephStore. Install `alephclient`")
    if uri.startswith("lake+"):
        try:
            from ftmq.store.lake import LakeStore

            uri = str(uri)[5:]
            return LakeStore(uri=uri, dataset=dataset, linker=linker)
        except ImportError:
            raise ImportError("Can not load LakeStore. Install `[lake]` dependencies")
    if uri.startswith("fragments+"):
        uri = str(uri)[10:]
        raise NotImplementedError(uri)
    raise NotImplementedError(uri)


__all__ = [
    "get_store",
    "Store",
    "View",
    "MemoryStore",
]



from datetime import datetime
from typing import Literal, TypeVar

from anystore.io import logged_items
from anystore.types import HttpUrlStr, SDict
from followthemoney.dataset import DataPublisher
from followthemoney.dataset.dataset import DatasetModel as _DatasetModel
from pydantic import AnyUrl
from rigour.mime.types import FTM

from ftmq.model.mixins import BaseModel
from ftmq.model.stats import DatasetStats
from ftmq.types import Entities
from ftmq.util import DEFAULT_DATASET

ContentType = Literal["documents", "structured", "mixed"]

D = TypeVar("D", bound="Dataset")


class Dataset(BaseModel, _DatasetModel):
    prefix: str | None = None
    maintainer: DataPublisher | None = None
    stats: DatasetStats = DatasetStats()
    git_repo: AnyUrl | None = None
    content_type: ContentType | None = None
    uri: str | None = None

    def iterate(self) -> Entities:
        from ftmq.io import smart_read_proxies  # FIXME

        for resource in self.resources:
            if resource.mime_type == FTM and resource.url:
                yield from logged_items(
                    smart_read_proxies(str(resource.url)),
                    "Read",
                    1_000,
                    uri=resource.url,
                    item_name="Proxy",
                    dataset=self.name,
                )

    def apply_stats(self, stats: DatasetStats) -> None:
        self.entity_count = stats.entity_count
        self.thing_count = stats.things.total
        self.stats = stats


def ensure_dataset(data: SDict | Dataset) -> Dataset:
    if isinstance(data, Dataset):
        return data
    return Dataset(**data)


class Catalog(BaseModel):
    name: str = "catalog"
    title: str = "Catalog"
    datasets: list[Dataset] = []
    updated_at: datetime | None = None
    description: str | None = None
    maintainer: DataPublisher | None = None
    publisher: DataPublisher | None = None
    url: HttpUrlStr | None = None
    uri: str | None = None
    logo_url: HttpUrlStr | None = None
    git_repo: AnyUrl | None = None

    def iterate(self) -> Entities:
        for dataset in self.datasets:
            yield from dataset.iterate()

    @property
    def names(self) -> set[str]:
        """Get the names of all datasets in the catalog."""
        return {d.name for d in self.datasets}


def make_dataset(name: str = DEFAULT_DATASET, cls: type[D] = Dataset, **kwargs) -> D:
    kwargs["title"] = kwargs.pop("title", name)
    return cls(name=name, **kwargs)
